/*******************************************************************************
 * Copyright (c) 2008 IGA Tosiki, NTT DATA BUSINESS BRAINS Corp.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    IGA Tosiki (NTT DATA BUSINESS BRAINS Corp.) - initial API and implementation
 *******************************************************************************/
/*
 * blanco Framework
 * Copyright (C) 2008 NTT DATA BUSINESS BRAINS CORPORATION
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 */
package blanco.html.parser.util;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StringWriter;
import java.util.Map;

import org.xml.sax.SAXException;

import blanco.html.parser.BlancoHtmlParser;
import blanco.html.parser.BlancoHtmlParserFactory;
import blanco.html.parser.helper.BlancoHtmlDecideEncodingContentHandler;
import blanco.xml.bind.BlancoXmlBindingUtil;
import blanco.xml.bind.BlancoXmlUnmarshaller;
import blanco.xml.bind.valueobject.BlancoXmlDocument;
import blanco.xml.bind.valueobject.BlancoXmlElement;

/**
 * HTMLp[T[̂߂̃[eBeBNXB
 * 
 * @author IGA Tosiki
 */
public class BlancoHtmlParserUtil {
    protected static final Map<String, String> fMapName2Value = BlancoHtmlCharacterEntitySetUtil
            .getName2ValueMap();

    protected static final Map<String, String> fMapValue2Name = BlancoHtmlCharacterEntitySetUtil
            .getValue2NameMap();

    /**
     * vf̂Ȃ̃eLXgɂăzCgXy[Xǂ𔻒肵܂B
     * 
     * szCgXy[XɊ܂߂Ĕf܂B
     * 
     * @param cRead
     *            肵B
     * @return zCgXy[XȂtrueB
     */
    public static boolean isWhiteSpace(final char cRead) {
        switch (cRead) {
        case ' ':// 0x20: SPACE
        case '\t': // 0x09: TAB
        case '\f':// 0x0c: FF
        case '\r':// 0x0d (line break)
        case '\n': // 0x0a (line break)
            return true;
        default:
            return false;
        }
    }

    /**
     * HTML̕R[hȂ܂B
     * 
     * @param argInputHtml
     *            oCgz񉻂ꂽ HTMLB
     * @return 茋ʂ̕GR[fBOB
     * @throws IOException
     *             o͗OꍇB
     * @throws SAXException
     */
    public static String decideEncoding(final byte[] argInputHtml)
            throws IOException, SAXException {
        final BlancoHtmlParser encodingParser = BlancoHtmlParserFactory
                .getInstance();
        final BlancoHtmlDecideEncodingContentHandler encodingHandler = new BlancoHtmlDecideEncodingContentHandler();
        encodingParser.setHandler(encodingHandler);

        encodingParser.parse(new BufferedReader(new InputStreamReader(
                new ByteArrayInputStream(argInputHtml), "ISO8859_1")));
        if (encodingHandler.getEncoding() != null) {
            // TODO JavaT|[gȂR[h^ꂽꍇ̗OB
            // TODO Shift_JIS Ɨ Windows-31J ɓǂݑւAȂǂ̂ŽKvB
            // TODO {_IɁAHTML̃GR[fBO ۂ
            // JavavO~Oňۂ̃GR[fBOƂ͕člKv悤ɎvB
            return encodingHandler.getEncoding();
        } else {
            // Ȃ̂Ō߂B
            // TODO zIɂ́A̕GR[fBOœĂĂ݂āAƂ ?
            // Ȃ̗̂pAȂǂ̏ʓr̂ǂƍlB
            return "Windows-31J";
        }
    }

    /**
     * QƂfR[h܂B
     * 
     * <UL>
     * <LI>^ꂽɂāAQƂ̃fR[h{\ł΂{܂B
     * <LI>{s\ł΁A^ꂽ̂܂ܖ߂܂B
     * <LI>fR[hɂ XML n𗘗p܂BXML  HTML Ŏ኱̑͂܂A͍l܂B
     * </UL>
     * 
     * @param argToken
     * @return
     */
    public static String decodeCharReferenceToken(final String argToken) {
        if (fMapName2Value.get(argToken.substring(1, argToken.length() - 1)) != null) {
            return String.valueOf((char) Integer.parseInt(fMapName2Value
                    .get(argToken.substring(1, argToken.length() - 1))));
        }

        final String xml = "<root>" + argToken + "</root>";
        try {
            final BlancoXmlDocument document = new BlancoXmlUnmarshaller()
                    .unmarshal(new ByteArrayInputStream(xml.getBytes()));
            final BlancoXmlElement element = BlancoXmlBindingUtil
                    .getDocumentElement(document);
            return BlancoXmlBindingUtil.getTextContent(element);
        } catch (IllegalArgumentException e) {
            // XML ƂĂ̕QƃfR[h͎{s\łB
            // A͋CɂɁA^ꂽ̂܂ܖ߂܂B
            // ȂÃ[gʉ߂ۂɁAXML nɂĂ͕WG[o͂ɕ񂪏o͂ꍇ܂B͖ĂB
            return argToken;
        }
    }

    public static String decodeCharReference(final String argInput) {
        final StringWriter writer = new StringWriter();
        for (int index = 0; index < argInput.length(); index++) {
            final char look = argInput.charAt(index);
            if (look != '&') {
                writer.append(look);
            } else {
                // QƔȂ܂B
                index++;
                final StringWriter innerWriter = new StringWriter();
                innerLoop: for (; index < argInput.length(); index++) {
                    final char lookInner = argInput.charAt(index);
                    switch (lookInner) {
                    case ';':
                        // I[B
                        break innerLoop;
                    case ' ':
                        // sBEB
                        innerWriter.append(lookInner);
                        break innerLoop;
                    default:
                        innerWriter.append(lookInner);
                        // TODO u[NČ̕KvB
                        break;
                    }
                }
                writer.write(BlancoHtmlParserUtil.decodeCharReferenceToken("&"
                        + innerWriter.toString() + ";"));
            }
        }
        writer.flush();
        return writer.toString();
    }

    /**
     * QƂGR[h܂B
     * 
     * @param argInput
     * @return
     */
    public static String encodeCharReference(final String argInput) {
        final StringWriter writer = new StringWriter();
        for (int index = 0; index < argInput.length(); index++) {
            final char look = argInput.charAt(index);
            if (fMapValue2Name.get(String.valueOf((int) look)) != null) {
                writer.write("&"
                        + fMapValue2Name.get(String.valueOf((int) look)) + ";");
            } else {
                writer.write(look);
            }
        }
        writer.flush();
        return writer.toString();
    }

    /**
     * ̕QƂfR[h܂B
     * 
     * @param argToken
     * @return
     */
    public static String decodeAttributeCharReference(final String argInput) {
        final StringWriter writer = new StringWriter();
        for (int index = 0; index < argInput.length(); index++) {
            final char look = argInput.charAt(index);
            if (look != '&') {
                writer.append(look);
            } else {
                // QƔȂ܂B
                index++;
                StringWriter innerWriter = new StringWriter();
                innerLoop: for (; index < argInput.length(); index++) {
                    final char lookInner = argInput.charAt(index);
                    switch (lookInner) {
                    case ';':
                        // Qƕ̏I[B
                        // HTML ̑̏ꍇɂ &quot; ̂ݔ肷B
                        if (innerWriter.toString().equals("quot")) {
                            writer.write('"');
                        } else {
                            writer.write(BlancoHtmlParserUtil
                                    .decodeCharReferenceToken("&"
                                            + innerWriter.toString() + ";"));
                        }
                        innerWriter = new StringWriter();
                        break innerLoop;
                    case ' ':
                    case '=':
                        // sBEB
                        writer.write("&" + innerWriter.toString() + lookInner);
                        innerWriter = new StringWriter();
                        break innerLoop;
                    default:
                        innerWriter.append(lookInner);
                        break;
                    }
                }
                if (innerWriter.toString().length() > 0) {
                    writer.write("&" + innerWriter.toString());
                }
            }
        }
        writer.flush();
        return writer.toString();
    }

    /**
     * ̕QƂGR[h܂B
     * 
     * @param argInput
     * @return
     */
    public static String encodeAttributeCharReference(final String argInput) {
        final StringWriter writer = new StringWriter();
        for (int index = 0; index < argInput.length(); index++) {
            final char look = argInput.charAt(index);
            if (look == '"') {
                writer.write("&quot;");
            } else {
                writer.write(look);
            }
        }
        writer.flush();
        return writer.toString();
    }
}
